import asyncio

import aiodbm


async def main():
    async with aiodbm.open("data.dbm", "c") as db:  # opening/creating database
        await db.set("alpha", "green")  # creating new key alpha with value green
        value = await db.get("alpha")  # fetching value for key alpha
        print(value)
        await db.delete("alpha")  # delete key alpha


asyncio.run(main())



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "aiodbm"
copyright = "2023, Erik Kalkoken"
author = "Erik Kalkoken"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.autodoc", "sphinx.ext.napoleon"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for autodoc -----------------------------------------------------
autodoc_default_options = {
    "members": True,
    "member-order": "alphabetical",
    "undoc-members": False,
    "exclude-members": "__weakref__, __init__",
}
autoclass_content = "class"
autodoc_class_signature = "separated"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_theme_options = {
    "description": "An AsyncIO bridge for DBM",
    "fixed_sidebar": True,
    "badge_branch": "master",
    "github_button": True,
    "github_user": "ErikKalkoken",
    "github_repo": "aiodbm",
    "show_powered_by": False,
    "sidebar_collapse": False,
    "extra_nav_links": {
        "Report Issues": "https://github.com/ErikKalkoken/aiodbm/issues",
    },
}

html_sidebars = {
    "**": [
        "about.html",
        "navigation.html",
        "relations.html",
        "searchbox.html",
    ],
}

html_static_path = ["_static"]
html_css_files = ["custom.css"]



"""Core logic for an async DBM proxy."""

import asyncio
import dbm
import logging
import queue
import threading
from functools import partial
from pathlib import Path
from typing import Any, Callable, Generator, List, NamedTuple, Optional, Union

logger = logging.getLogger("aiodbm")


class _Message(NamedTuple):
    """A queue message."""

    future: Optional[asyncio.Future]
    func: Optional[Callable]
    is_stop_signal: bool = False

    @property
    def future_strict(self) -> asyncio.Future:
        """Return future only if it exists, else raise exception."""
        if self.future is None:
            raise ValueError("No future")
        return self.future

    @property
    def func_strict(self) -> Callable:
        """Return function only if it exists, else raise exception."""
        if self.func is None:
            raise ValueError("No func")
        return self.func

    def __str__(self) -> str:
        return str(self.func)

    @classmethod
    def create_stop_signal(cls) -> "_Message":
        return cls(None, None, is_stop_signal=True)


class Database(threading.Thread):
    """A proxy for a DBM database."""

    def __init__(self, connector: Callable) -> None:
        super().__init__()
        self._db = None
        self._connector = connector
        self._message_queue = queue.Queue()

    def __await__(self) -> Generator[Any, None, "Database"]:
        return self._connect().__await__()

    async def __aenter__(self) -> "Database":
        return await self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    @property
    def _db_strict(self):
        """Return current database if one is active, else raise exception."""
        if self._db is None:
            raise ValueError("no active database")

        return self._db

    @property
    def _dbm_type_name(self) -> str:
        """Return name of the currently active DBM variant."""
        return type(self._db_strict).__name__

    @property
    def is_gdbm(self) -> bool:
        """Return True if this is a GDBM database."""
        return self._dbm_type_name == "gdbm"

    async def _connect(self) -> "Database":
        """Connect to the actual DBM database."""
        if self._db is not None:
            raise RuntimeError("Already connected")

        self.start()
        try:
            future = asyncio.get_running_loop().create_future()
            self._message_queue.put_nowait(_Message(future, self._connector))
            self._db = await future
        except Exception:
            self._db = None
            self._stop_runner()
            raise

        return self

    def run(self) -> None:
        """
        Execute function calls on a separate thread.

        :meta private:
        """

        while True:  # Continues running until stop signal is received
            message: _Message = self._message_queue.get()
            if message.is_stop_signal:
                break

            logger.debug("executing %s", message)
            try:
                result = message.func_strict()
            except BaseException as ex:
                logger.debug("returning exception %s", ex)

                def set_exception(fut, e):
                    if not fut.done():
                        fut.set_exception(e)

                message.future_strict.get_loop().call_soon_threadsafe(
                    set_exception, message.future_strict, ex
                )
            else:
                logger.debug("operation %s completed", message)

                def set_result(fut, result):
                    if not fut.done():
                        fut.set_result(result)

                message.future_strict.get_loop().call_soon_threadsafe(
                    set_result, message.future_strict, result
                )

    def _stop_runner(self):
        """Stop the thread runner."""
        stop_signal = _Message.create_stop_signal()
        self._message_queue.put_nowait(stop_signal)

    async def _execute(self, fn, *args, **kwargs) -> Any:
        """Queue a function with the given arguments for execution in the runner."""
        if self._db is None:
            raise ValueError("Database closed")

        func = partial(fn, *args, **kwargs)
        future = asyncio.get_running_loop().create_future()

        self._message_queue.put_nowait(_Message(future, func))

        return await future

    # DBM API

    async def close(self) -> None:
        """Complete queued operations and close the database."""

        if self._db is None:
            return

        try:
            await self._execute(self._db_strict.close)
        except Exception:
            logger.warning("exception occurred while closing database")
            raise
        finally:
            self._db = None
            self._stop_runner()

    async def delete(self, key: Union[str, bytes]) -> None:
        """Delete given key."""

        def _func():
            try:
                del self._db_strict[key]
            except KeyError as ex:
                raise KeyError(f"Key {key} does not exist") from ex

        return await self._execute(_func)

    async def exists(self, key: Union[str, bytes]) -> bool:
        """Return True when the given key exists, else False."""

        def _func():
            return key in self._db_strict

        return await self._execute(_func)

    async def get(
        self, key: Union[str, bytes], default: Optional[bytes] = None
    ) -> Optional[bytes]:
        """Get the value of key. If the key does not exist, return default."""

        return await self._execute(self._db_strict.get, key, default)

    async def keys(self) -> List[bytes]:
        """Return existing keys."""

        return await self._execute(self._db_strict.keys)

    async def set(self, key: Union[str, bytes], value: Union[str, bytes]) -> None:
        """Set key to hold the value.
        If key already holds a value, it is overwritten.
        """

        def _set():
            self._db_strict[key] = value

        await self._execute(_set)

    async def setdefault(self, key: Union[str, bytes], default: bytes) -> bytes:
        """Set key to hold the default value, if it does not yet exist.
        Or return current value of existing key.
        """

        return await self._execute(self._db_strict.setdefault, key, default)

    # GDBM only API

    async def firstkey(self) -> bytes:
        """Return the first key for looping over all keys. GDBM only."""

        return await self._execute(self._db_strict.firstkey)

    async def nextkey(self, key: Union[str, bytes]) -> Optional[bytes]:
        """Return the next key, when looping over all keys.
        Or return None, when the end of the loop has been reached.
        GDBM only.
        """

        return await self._execute(self._db_strict.nextkey, key)

    async def reorganize(self) -> None:
        """Reorganize the database. GDBM only."""

        await self._execute(self._db_strict.reorganize)

    async def sync(self) -> None:
        """When the database has been opened in fast mode,
        this method forces any unwritten data to be written to the disk.
        GDBM only.
        """

        await self._execute(self._db_strict.sync)


def open(file: Union[str, Path], *args, **kwargs) -> Database:
    """Create and return a proxy to the DBM database.

    Example:

    .. code-block:: Python

        async with open("example.dbm", "c") as db:
            ...

    Args:
        file: filename for the DBM database

    Returns:
        DBM database proxy object
    """

    def connector():
        filepath = str(file)
        return dbm.open(filepath, *args, **kwargs)

    return Database(connector)


async def whichdb(filename: Union[str, Path]) -> Optional[str]:
    """Return the name of the DBM implementation,
    which can be used for opening the given file.

    Or if the file can not be read return None.
    Or if the file is not in a known DBM format, return an empty string.
    """

    filename_str = str(filename)
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, dbm.whichdb, filename_str)



import asyncio
import dbm
import shutil
import sys
import tempfile
import time
import unittest
from pathlib import Path
from unittest.mock import Mock

import aiodbm
from aiodbm.core import _Message

python_version = f"{sys.version_info.major}{sys.version_info.minor}"


class TestMessage(unittest.IsolatedAsyncioTestCase):
    async def test_str(self):
        def alpha():
            pass

        # given
        future = asyncio.get_running_loop().create_future()
        message = _Message(future, alpha)
        # then
        self.assertIn("alpha", str(message))

    def test_can_create_stop_signal(self):
        # when
        message = _Message.create_stop_signal()
        # then
        self.assertTrue(message.is_stop_signal)

    def test_future_strict_raises_error_when_no_future(self):
        # given
        message = _Message.create_stop_signal()
        # when/then
        with self.assertRaises(ValueError):
            message.future_strict

    def test_func_strict_raises_error_when_no_func(self):
        # given
        message = _Message.create_stop_signal()
        # when/then
        with self.assertRaises(ValueError):
            message.func_strict


class DbmAsyncioTestCase(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.temp_dir = Path(tempfile.mkdtemp())
        self.data_path = str(self.temp_dir / "data.dat")

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir, ignore_errors=True)


class TestDbmFunctions(DbmAsyncioTestCase):
    async def test_whichdbm_should_return_name(self):
        # given
        async with aiodbm.open(self.data_path, "c") as db:
            await db.set("alpha", "green")

        # when
        result = await aiodbm.whichdb(self.data_path)
        # then
        self.assertIn(result, ["dbm.gnu", "dbm.ndbm", "dbm.dumb"])


class TestDatabase(unittest.TestCase):
    def test_should_raise_error_when_no_database_configured(self):
        # given
        db = aiodbm.Database(Mock())
        # when/then
        with self.assertRaises(ValueError):
            db._db_strict


class TestDatabaseAsync(DbmAsyncioTestCase):
    async def test_get_should_return_none_when_key_does_not_exists(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when
            result = await db.get("alpha")
            # then
            self.assertIsNone(result)

    async def test_get_should_return_default_when_key_does_not_exists(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when
            result = await db.get("alpha", b"yellow")
            # then
            self.assertEqual(result, b"yellow")

    async def test_close_should_close_the_db(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when/then
            await db.close()

    async def test_delete_should_remove_entry(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # given
            await db.set("alpha", "blue")
            # when
            await db.delete("alpha")
            # then
            self.assertFalse(await db.exists("alpha"))

    async def test_delete_should_raise_error_when_key_does_not_exist(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when/then
            with self.assertRaises(KeyError):
                await db.delete("alpha")

    async def test_exists_returns_true_when_key_exists(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # given
            await db.set("alpha", "blue")
            # when/then
            self.assertTrue(await db.exists("alpha"))

    async def test_exists_returns_false_when_key_does_not_exists(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when/then
            self.assertFalse(await db.exists("alpha"))

    async def test_keys_should_return_existing_keys(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # given
            await db.set("alpha", "blue")
            await db.set("bravo", "red")
            # when
            result = await db.keys()
            # then
            self.assertSetEqual(set(result), {b"bravo", b"alpha"})

    async def test_keys_should_return_empty_list_when_no_keys(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when
            result = await db.keys()
            # then
            self.assertListEqual(result, [])

    async def test_set_should_create_new_entry(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when
            await db.set("alpha", "blue")
            # then
            data = await db.get("alpha")
        self.assertEqual(data, b"blue")

    async def test_set_should_overwrite_existing_entry(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # given
            await db.set("alpha", "blue")
            # when
            await db.set("alpha", "yellow")
            # then
            data = await db.get("alpha")
            self.assertEqual(data, b"yellow")

    async def test_set_should_add_a_new_entry(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when
            await db.set("alpha", "blue")
            # then
            data = await db.get("alpha")
            self.assertEqual(data, b"blue")

    async def test_set_should_raise_error_on_wrong_types(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when/then
            with self.assertRaises(TypeError):
                await db.set("alpha", 1)  # type: ignore

    async def test_setdefault_should_add_new_entry_when_key_does_not_exit(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when
            result = await db.setdefault("alpha", b"blue")
            # then
            self.assertEqual(result, b"blue")

    async def test_setdefault_should_return_existing_value_when_key_exists(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # given
            await db.set("alpha", "green")
            # when
            result = await db.setdefault("alpha", b"blue")
            # then
            self.assertEqual(result, b"green")

    async def test_should_stop_thread_after_leaving_context(self):
        async with aiodbm.open(self.data_path, "c") as db:
            pass
        # then
        time.sleep(1)
        self.assertFalse(db.is_alive())

    async def test_should_stop_thread_when_closing(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when
            await db.close()
            # then
            time.sleep(1)
            self.assertFalse(db.is_alive())

    async def test_can_open_without_context_manager(self):
        db = await aiodbm.open(self.data_path, "c")
        await db.get("dummy")
        await db.close()

    async def test_open_raises_exception_when_opening_fails(self):
        with self.assertRaises(dbm.error):
            await aiodbm.open(self.data_path, "r")

    async def test_can_call_close_multiple_times(self):
        # given
        db = await aiodbm.open(self.data_path, "c")
        await db.close()
        # when/then
        await db.close()

    async def test_should_raise_error_when_trying_to_connect_again(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when/then
            with self.assertRaises(RuntimeError):
                await db._connect()


@unittest.skipIf(python_version in ["38", "39"], reason="Unsupported Python")
class TestGdbmFunctions(DbmAsyncioTestCase):
    async def test_firstkey_should_return_first_key(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # given
            await db.set("alpha", "green")
            # when
            result = await db.firstkey()
            # then
            self.assertEqual(result, b"alpha")

    async def test_can_loop_though_all_keys(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # given
            await db.set("alpha", "green")
            await db.set("bravo", "green")
            await db.set("charlie", "green")
            # when
            keys = set()
            key = await db.firstkey()
            while key is not None:
                keys.add(key)
                key = await db.nextkey(key)
            # then
            self.assertSetEqual(keys, {b"alpha", b"bravo", b"charlie"})

    async def test_can_reorganize(self):
        async with aiodbm.open(self.data_path, "c") as db:
            # when/then
            await db.reorganize()

    async def test_can_sync(self):
        async with aiodbm.open(self.data_path, "cf") as db:
            # when/then
            await db.set("alpha", "green")
            await db.sync()

    async def test_can_detect_gdbm(self):
        # given
        async with aiodbm.open(self.data_path, "c") as db:
            await db.set("alpha", "green")
        result = await aiodbm.whichdb(self.data_path)
        assert result == "dbm.gnu"
        # when/then
        async with aiodbm.open(self.data_path, "c") as db:
            self.assertTrue(db.is_gdbm)



"""Measure and compare throughput between normal DBM and asyncio version."""

import asyncio
import dbm
import random
import string
import tempfile
import time
from pathlib import Path

import aiodbm

ITEMS_AMOUNT = 10_000
OBJECT_SIZE = 256


def random_string(length: int) -> str:
    return "".join(random.choices(string.ascii_uppercase + string.digits, k=length))


def generate_objs() -> dict:
    objs = {
        f"item-{num:010}": random_string(OBJECT_SIZE)
        for num in range(1, ITEMS_AMOUNT + 1)
    }
    return objs


def run_normal():
    with tempfile.TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "measurement_1.dbm"
        objs = generate_objs()
        objs_2 = {}
        with dbm.open(str(path), "c") as db:
            start = time.perf_counter()
            for key, value in objs.items():
                db[key] = value
            for key in objs.keys():
                objs_2[key] = db[key]
            end = time.perf_counter()

    duration = end - start
    return duration


async def run_async():
    with tempfile.TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "measurement_2.dbm"
        objs = generate_objs()
        objs_2 = {}
        async with aiodbm.open(str(path), "c") as db:
            start = time.perf_counter()
            for key, value in objs.items():
                await db.set(key, value)
            for key in objs.keys():
                objs_2[key] = await db.get(key)

            end = time.perf_counter()

    duration = end - start
    return duration


def main():
    throughput_normal = ITEMS_AMOUNT / run_normal()
    print(f"Throughput measurement results for {ITEMS_AMOUNT:,} items:")
    print(f"Normal: {throughput_normal:,.1f} items / sec")

    throughput_async = ITEMS_AMOUNT / asyncio.run(run_async())
    print(f"Async: {throughput_async:,.1f} items / sec")
    factor = int(throughput_async / throughput_normal * 100)
    print(f"Throughput of async version is {factor}% of normal")


if __name__ == "__main__":
    main()
